# coding: utf-8
from dataclasses import dataclass
from urllib.parse import urlsplit

import pytest

from lektor.utils import build_url
from lektor.utils import is_path_child_of
from lektor.utils import join_path
from lektor.utils import magic_split_ext
from lektor.utils import make_relative_url
from lektor.utils import parse_path
from lektor.utils import secure_url
from lektor.utils import slugify
from lektor.utils import Url


def test_join_path():

    assert join_path("a", "b") == "a/b"
    assert join_path("/a", "b") == "/a/b"
    assert join_path("a@b", "c") == "a@b/c"
    assert join_path("a@b", "") == "a@b"
    assert join_path("a@b", "@c") == "a@c"
    assert join_path("a@b/c", "a@b") == "a/a@b"

    assert join_path("blog@archive", "2015") == "blog@archive/2015"
    assert join_path("blog@archive/2015", "..") == "blog@archive"
    assert join_path("blog@archive/2015", "@archive") == "blog@archive"
    assert join_path("blog@archive", "..") == "blog"
    assert join_path("blog@archive", ".") == "blog@archive"
    assert join_path("blog@archive", "") == "blog@archive"

    # special behavior: parent of pagination paths is always the actual
    # page parent.
    assert join_path("/blog@1", "..") == "/"
    assert join_path("/blog@2", "..") == "/"

    # But joins on the same level keep the path
    assert join_path("/blog@1", ".") == "/blog@1"
    assert join_path("/blog@2", ".") == "/blog@2"
    assert join_path("/blog@1", "") == "/blog@1"
    assert join_path("/blog@2", "") == "/blog@2"


def test_is_path_child_of():

    assert not is_path_child_of("a/b", "a/b")
    assert is_path_child_of("a/b", "a/b", strict=False)
    assert is_path_child_of("a/b/c", "a")
    assert not is_path_child_of("a/b/c", "b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert is_path_child_of("a/b@foo", "a/b@foo", strict=False)
    assert not is_path_child_of("a/b@foo/bar", "a/c@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/c")
    assert is_path_child_of("a/b@foo", "a/b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/b@bar")


def test_magic_split_ext():

    assert magic_split_ext("wow") == ("wow", "")
    assert magic_split_ext("aaa.jpg") == ("aaa", "jpg")
    assert magic_split_ext("aaa. jpg") == ("aaa. jpg", "")
    assert magic_split_ext("aaa.j pg") == ("aaa.j pg", "")
    assert magic_split_ext("aaa.j pg", ext_check=False) == ("aaa", "j pg")


def test_slugify():

    assert slugify("w o w") == "w-o-w"
    assert slugify("Șö prĕtty") == "so-pretty"
    assert slugify("im age.jpg") == "im-age.jpg"
    assert slugify("slashed/slug") == "slashed/slug"


@dataclass
class SampleUrl:
    uri: str
    iri: str

    @property
    def split_uri(self):
        return urlsplit(self.uri)

    @property
    def split_iri(self):
        return urlsplit(self.iri)


SAMPLE_URLS = [
    SampleUrl("https://example.org/foo", "https://example.org/foo"),
    SampleUrl("https://example.org:8001/f%C3%BC", "https://example.org:8001/fü"),
    SampleUrl(
        "https://xn--wgv71a119e.idn.icann.org/%E5%A4%A7",
        "https://日本語.idn.icann.org/大",
    ),
    SampleUrl("/?q=sch%C3%B6n#gru%C3%9F", "/?q=schön#gruß"),
]


@pytest.fixture(params=SAMPLE_URLS, ids=lambda sample: sample.uri)
def sample_url(request):
    sample_url = request.param
    # sanity checks
    assert sample_url.split_uri.scheme == sample_url.split_iri.scheme
    assert sample_url.split_uri.port == sample_url.split_iri.port
    return sample_url


def test_Url_str(sample_url):
    assert str(Url(sample_url.iri)) == sample_url.iri
    assert str(Url(sample_url.uri)) == sample_url.uri


def test_Url_ascii_url(sample_url):
    assert Url(sample_url.iri).ascii_url == sample_url.uri
    assert Url(sample_url.uri).ascii_url == sample_url.uri


def test_Url_ascii_host(sample_url):
    assert Url(sample_url.iri).ascii_host == sample_url.split_uri.hostname
    assert Url(sample_url.uri).ascii_host == sample_url.split_uri.hostname


def test_Url_scheme(sample_url):
    assert Url(sample_url.iri).scheme == sample_url.split_uri.scheme
    assert Url(sample_url.uri).scheme == sample_url.split_uri.scheme


def test_Url_host(sample_url):
    assert Url(sample_url.iri).host == sample_url.split_iri.hostname
    assert Url(sample_url.uri).host == sample_url.split_iri.hostname


def test_Url_port(sample_url):
    assert Url(sample_url.iri).port == sample_url.split_uri.port
    assert Url(sample_url.uri).port == sample_url.split_uri.port


def test_Url_path(sample_url):
    assert Url(sample_url.iri).path == sample_url.split_iri.path
    assert Url(sample_url.uri).path == sample_url.split_iri.path


def test_Url_query(sample_url):
    try:
        assert Url(sample_url.iri).query == sample_url.split_iri.query
        assert Url(sample_url.uri).query == sample_url.split_iri.query
    except AssertionError:
        # This is the behavior prior to Lektor 3.4.x
        assert Url(sample_url.iri).query == sample_url.split_iri.query
        assert Url(sample_url.uri).query == sample_url.split_uri.query
        pytest.xfail("Url.query is weird in Lektor<3.4")


def test_Url_anchor(sample_url):
    assert Url(sample_url.iri).anchor == sample_url.split_iri.fragment
    assert Url(sample_url.uri).anchor == sample_url.split_iri.fragment


@pytest.mark.parametrize(
    "url, expected",
    [
        ("https://user:pw@example.org/p", "https://user@example.org/p"),
        ("https://user:pw@example.org:8000", "https://user@example.org:8000"),
        ("https://user@example.org/b", "https://user@example.org/b"),
    ],
)
def test_secure_url(url, expected):
    assert secure_url(url) == expected


def test_url_builder():

    assert build_url([]) == "/"
    assert build_url(["a", "b/c"]) == "/a/b/c/"
    assert build_url(["a", "b/c"], trailing_slash=False) == "/a/b/c"
    assert build_url(["a", "b/c.html"]) == "/a/b/c.html"
    assert build_url(["a", "b/c.html"], trailing_slash=True) == "/a/b/c.html/"
    assert build_url(["a", None, "b", "", "c"]) == "/a/b/c/"


def test_parse_path():

    assert parse_path("") == []
    assert parse_path("/") == []
    assert parse_path("/foo") == ["foo"]
    assert parse_path("/foo/") == ["foo"]
    assert parse_path("/foo/bar") == ["foo", "bar"]
    assert parse_path("/foo/bar/") == ["foo", "bar"]
    assert parse_path("/foo/bar/../stuff") == ["foo", "bar", "stuff"]


@pytest.mark.parametrize(
    "source, target, expected",
    [
        ("/", "./a/", "a/"),
        ("/", "./a", "a"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("/fr/blog/2015/11/a/", "/fr/blog/", "../../../"),
        ("/fr/blog/2015/11/a.php", "/fr/blog/", "../../"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/", "../../../2016/"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/", "../2015/a/"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("/fr/blog/2015/11/a/", "/images/b.svg", "../../../../../images/b.svg"),
        ("/fr/blog/", "2015/11/", "2015/11/"),
        ("/fr/blog/x", "2015/11/", "2015/11/"),
        ("", "./a/", "a/"),
        ("", "./a", "a"),
        ("fr/blog/2015/11/a/", "fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("fr/blog/2015/11/a/", "fr/blog/", "../../../"),
        ("fr/blog/2015/11/a.php", "fr/blog/", "../../"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/", "../../../2016/"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("fr/blog/2016/", "fr/blog/2015/a/", "../2015/a/"),
        ("fr/blog/2016/", "fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("fr/blog/2015/11/a/", "images/b.svg", "../../../../../images/b.svg"),
        ("fr/blog/", "2015/11/", "../../2015/11/"),
        ("fr/blog/x", "2015/11/", "../../2015/11/"),
    ],
)
def test_make_relative_url(source, target, expected):
    assert make_relative_url(source, target) == expected


def test_make_relative_url_relative_source_absolute_target():
    with pytest.raises(ValueError):
        make_relative_url("rel/a/tive/", "/abs/o/lute")



import os
from pathlib import Path

import pytest

from lektor.packages import add_site
from lektor.packages import install_local_package
from lektor.pluginsystem import load_plugins


@pytest.mark.usefixtures("save_sys_path")
def test_install_local_package(tmp_path):
    package_root = os.fspath(tmp_path / "package_root")
    dummy_plugin_path = os.fspath(Path(__file__).parent / "setup_py-dummy-plugin")
    install_local_package(package_root, dummy_plugin_path)
    add_site(package_root)
    plugins = load_plugins()
    assert plugins["dummy-plugin"].name == "dummy"



from setuptools import setup

setup(
    name="lektor-dummy-plugin",
    description="setup.py test plugin",
    version="0.1a42",
    py_modules=["lektor_dummy_plugin"],
    entry_points={
        "lektor.plugins": [
            "dummy-plugin = lektor_dummy_plugin:DummyPlugin",
        ]
    },
)



import functools
import os
import shutil
import sys
import threading
import time
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Generator
from typing import Optional
from typing import Type

import py
import pytest
from watchdog.observers.api import BaseObserver
from watchdog.observers.polling import PollingObserver

from lektor import utils
from lektor.watcher import BasicWatcher
from lektor.watcher import watch
from lektor.watcher import Watcher


class BrokenObserver(PollingObserver):
    # The InotifyObserver, when it fails due to insufficient system
    # inotify resources, does not fail until an attempt is made to start it.
    def start(self):
        raise OSError("crapout")


class TestBasicWatcher:
    # pylint: disable=no-self-use

    @pytest.fixture
    def paths(self, tmp_path):
        return [str(tmp_path)]

    def test_creates_observer(self, paths):
        with BasicWatcher(paths) as watcher:
            assert isinstance(watcher.observer, BaseObserver)

    def test_default_observer_broken(self, paths, capsys):
        observer_classes = (BrokenObserver, PollingObserver)
        with BasicWatcher(paths, observer_classes=observer_classes) as watcher:
            assert watcher.observer.__class__ is PollingObserver
        assert "crapout" in capsys.readouterr().out

    def test_default_observer_is_polling(self, paths, capsys):
        observer_classes = (BrokenObserver, BrokenObserver)
        with pytest.raises(OSError, match=r"crapout"):
            with BasicWatcher(paths, observer_classes=observer_classes):
                pass
        assert capsys.readouterr() == ("", "")

    def test_perverse_usage(self, paths):
        # This exercises a bug which occurred when BasicWatcher was
        # called with repeated (failing) values in observer_classes.
        observer_classes = (BrokenObserver, BrokenObserver, PollingObserver)
        with BasicWatcher(paths, observer_classes=observer_classes) as watcher:
            assert isinstance(watcher.observer, BaseObserver)

    def test_raises_error_if_started_twice(self, paths):
        with BasicWatcher(paths) as watcher:
            with pytest.raises(RuntimeError, match="already started"):
                watcher.start()


@dataclass
class WatcherTest:
    watched_path: Path

    @contextmanager
    def __call__(
        self,
        observer_class: Optional[Type[BaseObserver]] = None,
        should_set_event: bool = True,
        timeout: float = 1.2,
    ) -> Generator[Path, None, None]:

        kwargs: dict[str, Any] = {}
        if observer_class is not None:
            kwargs.update(
                observer_classes=(observer_class,),
                observer_timeout=0.1,  # fast polling timer to speed tests
            )

        with BasicWatcher([os.fspath(self.watched_path)], **kwargs) as watcher:
            event = threading.Event()

            class WatcherWatcher(threading.Thread):
                # iterate watcher in a separate thread
                def run(self):
                    for _ in watcher:
                        event.set()

            WatcherWatcher(daemon=True).start()

            if sys.platform == "darwin":
                # The FSEventObserver (used on macOS) seems to send events for things that
                # happened before is was started.  Here, we wait a little bit for things to
                # start, then discard any pre-existing events.
                time.sleep(0.1)
                event.clear()

            yield self.watched_path

            if should_set_event:
                assert event.wait(timeout)
            else:
                assert not event.wait(timeout)


@pytest.fixture(
    params=[
        pytest.param(None, id="default observer"),
        PollingObserver,
    ]
)
def observer_class(request: pytest.FixtureRequest) -> bool:
    return request.param


@pytest.fixture
def watcher_test(tmp_path: Path) -> WatcherTest:
    watched_path = tmp_path / "watched_path"
    watched_path.mkdir()
    return WatcherTest(watched_path)


def test_watcher_test(watcher_test: WatcherTest) -> None:
    with watcher_test(should_set_event=False, timeout=0.2):
        pass


def test_BasicWatcher_sees_created_file(
    watcher_test: WatcherTest, observer_class: Optional[Type[BaseObserver]]
) -> None:
    with watcher_test(observer_class=observer_class) as watched_path:
        Path(watched_path, "created").touch()


def test_BasicWatcher_sees_deleted_file(
    watcher_test: WatcherTest, observer_class: Optional[Type[BaseObserver]]
) -> None:
    deleted_path = watcher_test.watched_path / "deleted"
    deleted_path.touch()

    with watcher_test(observer_class=observer_class):
        deleted_path.unlink()


def test_BasicWatcher_sees_modified_file(
    watcher_test: WatcherTest, observer_class: Optional[Type[BaseObserver]]
) -> None:
    modified_path = watcher_test.watched_path / "modified"
    modified_path.touch()

    with watcher_test(observer_class=observer_class):
        with modified_path.open("a") as fp:
            fp.write("addition")


def test_BasicWatcher_sees_file_moved_in(
    watcher_test: WatcherTest,
    observer_class: Optional[Type[BaseObserver]],
    tmp_path: Path,
) -> None:
    orig_path = tmp_path / "orig_path"
    orig_path.touch()
    final_path = watcher_test.watched_path / "final_path"

    with watcher_test(observer_class=observer_class):
        orig_path.rename(final_path)


def test_BasicWatcher_sees_file_moved_out(
    watcher_test: WatcherTest,
    observer_class: Optional[Type[BaseObserver]],
    tmp_path: Path,
) -> None:
    orig_path = watcher_test.watched_path / "orig_path"
    orig_path.touch()
    final_path = tmp_path / "final_path"

    with watcher_test(observer_class=observer_class):
        orig_path.rename(final_path)


def test_BasicWatcher_sees_deleted_directory(
    watcher_test: WatcherTest, observer_class: Optional[Type[BaseObserver]]
) -> None:
    # We only really care about deleted directories that contain at least a file.
    deleted_path = watcher_test.watched_path / "deleted"
    deleted_path.mkdir()
    watched_file = deleted_path / "file"
    watched_file.touch()

    with watcher_test(observer_class=observer_class):
        shutil.rmtree(deleted_path)


def test_BasicWatcher_sees_file_in_directory_moved_in(
    watcher_test: WatcherTest,
    observer_class: Optional[Type[BaseObserver]],
    tmp_path: Path,
) -> None:
    # We only really care about directories that contain at least a file.
    orig_dir_path = tmp_path / "orig_dir_path"
    orig_dir_path.mkdir()
    Path(orig_dir_path, "file").touch()
    final_dir_path = watcher_test.watched_path / "final_dir_path"

    with watcher_test(observer_class=observer_class):
        orig_dir_path.rename(final_dir_path)


def test_BasicWatcher_sees_directory_moved_out(
    watcher_test: WatcherTest,
    observer_class: Optional[Type[BaseObserver]],
    tmp_path: Path,
) -> None:
    # We only really care about directories that contain at least one file.
    orig_dir_path = watcher_test.watched_path / "orig_dir_path"
    orig_dir_path.mkdir()
    Path(orig_dir_path, "file").touch()
    final_dir_path = tmp_path / "final_dir_path"

    with watcher_test(observer_class=observer_class):
        orig_dir_path.rename(final_dir_path)


def test_BasicWatcher_ignores_opened_file(watcher_test: WatcherTest) -> None:
    file_path = watcher_test.watched_path / "file"
    file_path.touch()

    with watcher_test(should_set_event=False):
        with file_path.open() as fp:
            fp.read()


def test_is_interesting(env):
    # pylint: disable=no-member
    cache_dir = py.path.local(utils.get_cache_dir())
    build_dir = py.path.local("build")

    w = Watcher(env, str(build_dir))

    # This partial makes the testing code shorter
    is_interesting = functools.partial(w.is_interesting, 0, "generic")

    assert is_interesting("a.file")
    assert not is_interesting(".file")
    assert not is_interesting(str(cache_dir / "another.file"))
    assert not is_interesting(str(build_dir / "output.file"))

    w.output_path = None
    assert is_interesting(str(build_dir / "output.file"))


def test_watch(env, mocker):
    Watcher = mocker.patch("lektor.watcher.Watcher")
    event1 = mocker.sentinel.event1

    def events():
        yield event1
        raise KeyboardInterrupt()

    watcher = Watcher.return_value.__enter__.return_value
    watcher.__iter__.return_value = events()

    assert list(watch(env)) == [event1]
